import os

from config_joker.sources.source import Source, SourceResponse


class EnvironmentSource(Source):
    def get_value(self, key: str) -> SourceResponse:
        try:
            return SourceResponse(
                exists=True,
                value=os.environ[key]
            )
        except KeyError:
            return SourceResponse(
                exists=False,
                value=None
            )



from config_joker.sources.source import Source, SourceResponse
from typing import List, Any, Callable


class RequiredKeyNotFound(Exception):
    def __init__(self, key: str) -> None:
        super().__init__(f'Could not find the key "{key}"')


class Config:
    def __init__(self, sources: List[Source]) -> None:
        self._sources = sources

    def _get(self, key: str) -> SourceResponse:
        'Iters in the list of sources until it finds the key in one of them.'
        for source in self._sources:
            result = source.get_value(key=key)
            if result.exists:
                return result
        return result

    def required(self, key: str, value_type: Callable = None) -> Any:
        source_response = self._get(key=key)
        if not source_response.exists:
            raise RequiredKeyNotFound(key=key)
        result = source_response.value
        if value_type:
            result = value_type(result)
        return result

    def optional(self, key: str, value_type: Callable = None, default: Any = None) -> Any:
        source_response = self._get(key=key)
        if not source_response.exists:
            return default
        result = source_response.value
        if value_type:
            result = value_type(result)
        return result



from yaml import safe_load

from config_joker.sources.source import Source, SourceResponse
from config_joker.utils.parser import dict_extractor


class YamlFileSource(Source):
    def __init__(self, file_path: str) -> None:
        self._file_path = file_path
        self._data = safe_load(self._file_path)

    def get_value(self, key: str) -> SourceResponse:
        try:
            response = dict_extractor(path=key, data=self._data)
            return SourceResponse(
                exists=True,
                value=response
            )
        except KeyError:
            return SourceResponse(
                exists=False,
                value=None
            )



from typing import Any
from abc import ABC, abstractmethod

from dataclasses import dataclass


@dataclass
class SourceResponse:
    exists: bool
    value: Any


class Source(ABC):
    @abstractmethod
    def get_value(self, key: str) -> SourceResponse:
        'Get the value in the source using the key'

s = SourceResponse(**{"exists": True, "value": "1"})


from typing import Union, List, Any
import regex as re


def transform_splited_by_brackets(value: str) -> Union[str, int]:
    try:
        return int(value)
    except ValueError:
        return value


def flatten(lst):
    result = []
    for item in lst:
        if item=='':
            continue
        if isinstance(item, list):
            result.extend(flatten(item))
        else:
            result.append(item)
    return result


def path_parser(path: str) -> List[Union[str, int]]:
    dots_pattern = r"\."
    brackets_pattern = r"\[(.*?)\]"
    splited_by_brackets = [transform_splited_by_brackets(v) for v in re.split(brackets_pattern, path) if bool(v)]
    neasted_result = [re.split(dots_pattern, subpath) if isinstance(subpath, str) else subpath for subpath in splited_by_brackets]
    return flatten(neasted_result)


def dict_extractor(path: str, data: str) -> Any:
    parsed_path = path_parser(path=path)
    for key in parsed_path:
        data = data[key]
    return data